/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */
#include "CountWithSpectro.h"
#include "controllers/common/family/Families.h"
#include "controllers/common/datafile/NexusDataFile.h"
#include <boost/date_time/posix_time/posix_time.hpp>
#include <common/base/Date.h>

namespace d22special {

const std::string CountWithSpectro::TYPE = "count_with_spectro";
const int32 CountWithSpectro::XSIZE = 1044; //should be retrieved directly from the driver...
const int32 CountWithSpectro::YSIZE = 1044;

const int32 CountWithSpectro::USB_DELAY_MILLISEC = 50; //Empirical value obtained from tests

CountWithSpectro::CountWithSpectro(const std::string & name) :
		ExperimentController(name), controller::Stoppable(this) {

	setFamily(family::SAMPLE_ENVIRONMENT);

	registerFunction(TYPE);

	spectro.init(this, "spectro");
	count.init(this, "count");
	spectroSwitcher.init(this, "spectro_switcher");

	integrationTime.init(this, SAVE, "integration_time"); //in microseconds!
	numSpectroMeasures.init(this, SAVE, "num_spectro");
	numTotalMeasures.init(this, SAVE, "num_total_measures");
	actualPosition.init(this, NOSAVE, "actual_pos");
	timeType.init(this, SAVE, "time_type");
	countTime.init(this, SAVE, "count_time");
	scansToAverage.init(this, SAVE, "scans_average");

	xSize.init(this, SAVE, "x_size");
	ySize.init(this, SAVE, "y_size");
	zSize.init(this, SAVE, "z_size");

	subtitle.init(this, SAVE, "subtitle");
	useSwitcher.init(this, SAVE, "use_switcher");
	inSequencerMode.init(this, NOSAVE, "in_sequencer_mode");

	yDataArray.init(this, NOSAVE, "yData_array");
	initialTime.init(this, NOSAVE, "initial_time");
	measureDuration.init(this, NOSAVE, "measure_duration");

	yDataArraySecondChannel.init(this, NOSAVE, "yDataArray_secondChannel");

	//initialize pointers
	m_yDataArray = nullptr;
	m_yDataArraySecondChannel = nullptr;

	m_yDataOddPositionArray = nullptr;
	m_yDataEvenPositionArray = nullptr;
	m_yDataArrayToSave = nullptr;

}

CountWithSpectro::~CountWithSpectro() {
	delete[] m_yDataArray;
	delete[] m_yDataArraySecondChannel;

}

void CountWithSpectro::postConfiguration() {
	//	xSize = spectro->size();
	ySize = numSpectroMeasures();
	zSize = 1;

	registerRefresher(numSpectroMeasures, &CountWithSpectro::refreshNumSpectroMeasures, this);
	registerRefresher(useSwitcher, &CountWithSpectro::refreshUseSwitcher, this);
	registerUpdater(count->subtitle, &CountWithSpectro::updateSubtitle, this);
	registerProgression(count, &CountWithSpectro::updateProgression, this);

	initArrays();

	cleanArrays();
}

void CountWithSpectro::refreshNumSpectroMeasures(int32 number) {
	ySize = number; //This property is only used to save the data in nexus file

	//initialize here the dynamic properties
	if (useSwitcher()) {
		measureDuration.resize(number * 2);
		numTotalMeasures = number * 2;
	} else {
		measureDuration.resize(number);
		numTotalMeasures = number;
	}
}

void CountWithSpectro::refreshUseSwitcher(bool value) {
	if (value) {
		numTotalMeasures = numSpectroMeasures() * 2; //using switcher
	} else {
		numTotalMeasures = numSpectroMeasures();
	}
	measureDuration.resize(numTotalMeasures());
}

void CountWithSpectro::updateProgression() {
	//commandProgression = count->commandProgression();
}

void CountWithSpectro::updateSubtitle() {
	subtitle = count->subtitle();
}

void CountWithSpectro::start() {
	//to avoid problems of synchronization, get the xSize any time we do a measure
	xSize = spectro->size();
	spectro->scansToAverage = scansToAverage();

	if (spectro->isConnected()) {
		if (!useSwitcher()) {

			//Time already in microseconds!
			spectro->integrationTime = integrationTime();

			//check the total time of spectro measurement is smaller than the count time (comparison in seconds!)
			float64 spectroTime = integrationTime() * 0.000001 * numSpectroMeasures() * scansToAverage();
			float64 countTime = getCountTimeInSec();

			if (spectroTime <= countTime) {

				setupCountParams();

				initArrays();

				commandStatus.setRunning();
				spectro->mode = QE65000Controller::COUNT_SINGLE;

				//Launch count and spectro at the same time
				// Run the threads in parallel in a group
				boost::thread_group group;

				group.create_thread(boost::bind(&CountWithSpectro::startCount, this));
				group.create_thread(boost::bind(&CountWithSpectro::startSpectro, this));

				// Wait for the termination of the threads
				group.join_all();

				//Append data
				if (!inSequencerMode()) {
					saveSpectroData();
				}

				commandStatus.setIdle();
				spectro->mode = QE65000Controller::IDLE;
			}

		} else {

			//Suppose here we also keep the same rule as before (but now is double)
			spectroSwitcher->numberOfCycles = 1; //even if it is not clear, the number of cycles will be established by this controller
			spectroSwitcher->integrationTime = integrationTime();

			//check the total time of spectro measurement is smaller than the count time
			float64 spectroTime = integrationTime() * 0.000001 * numSpectroMeasures() * scansToAverage() * 2; //left and right measure
			float64 countTime = getCountTimeInSec();

			if (spectroTime <= countTime) {

				setupCountParams();

				//	initArrays();

				commandStatus.setRunning();
				spectro->mode = QE65000Controller::COUNT_SWITCHER;

				//Launch count and spectro at the same time
				// Run the threads in parallel in a group
				boost::thread_group group;

				group.create_thread(boost::bind(&CountWithSpectro::startCount, this));
				group.create_thread(boost::bind(&CountWithSpectro::startSwitcher, this));

				// Wait for the termination of the threads
				group.join_all();

				//Get data
				saveSwitcherSpectroData();
				commandStatus.setIdle();
				spectro->mode = QE65000Controller::IDLE;
			}
		}
		//clean arrays
		//initArrays();
	}
}

void CountWithSpectro::stop() {
	if (commandStatus.isRunning()) {
		count->stopParallel();
		spectro->stopParallel();
		spectroSwitcher->stopParallel();
	}
	commandStatus.setWarning();
}

void CountWithSpectro::startCount() {
	count->startCommand();
}

void CountWithSpectro::startSpectro() {
	common::Date startTime = common::Date();
	initialTime = startTime.toString("%H:%M:%S");
	common::Date formerTime = startTime;

	int32 sleepTime = calculateWaitingTimeInSec();

	for (int i = 0; i < numSpectroMeasures(); i++) {
		if (commandStatus.isRunning()) {

			common::Date time = common::Date();
			common::Duration duration = time - formerTime;
			measureDuration.set(i, duration.getMilliseconds());

			spectro->startCommand();

			int32 ySize = spectro->size();
			float64* tempYData = spectro->yData();
			for (int32 j = 0; j < ySize; j++) {
				m_yDataArray[ySize * i + j] = tempYData[j];
			}

			formerTime = time;
			sleep(sleepTime);
		}
	}

	// When using the controller with the samplesequencer we need to  store locally the spectrum data after each measure until its related count data is ready
	if (inSequencerMode()) {
		saveSpectroDataInShiftedMode();
	}
}

void CountWithSpectro::startSwitcher() {
	common::Date startTime = common::Date();
	initialTime = startTime.toString("%H:%M:%S");

	int32 sleepTime = calculateWaitingTimeInSec();

	common::Date initialTime = common::Date();

	for (int i = 0; i < numSpectroMeasures(); i++) {
		if (commandStatus.isRunning()) {

			spectroSwitcher->setInitialTime(initialTime);

			spectroSwitcher->startCommand();

			int32 ySize = spectro->size();
			float64* tempLeftYData = spectroSwitcher->yLeftData();
			for (int32 j = 0; j < ySize; j++) {
				m_yDataArray[ySize * i + j] = tempLeftYData[j];
			}

			float64* tempRightYData = spectroSwitcher->yRightData();
			for (int32 j = 0; j < ySize; j++) {
				m_yDataArraySecondChannel[ySize * i + j] = tempRightYData[j];
			}

			measureDuration.set((i * 2), spectroSwitcher->durationLeft());
			measureDuration.set((i * 2 + 1), spectroSwitcher->durationRight());

			initialTime = common::Date();
			sleep(sleepTime);
		}
	}

}

void CountWithSpectro::setupCountParams() {
	count->time.setpoint = countTime();
	count->timeType = timeType();
	count->subtitle = subtitle();
}

void CountWithSpectro::initArrays() {
	if (m_yDataArray!= NULL) {
		delete[] m_yDataArray;

	} else if (m_yDataArraySecondChannel) {
		delete[] m_yDataArraySecondChannel;
	}

	int32 totalSize = spectro->size() * numSpectroMeasures();
	m_yDataArray = new float64[totalSize];
	memset(m_yDataArray, 0, totalSize * sizeof(float64)); //set to zero
	yDataArray.update(m_yDataArray);
	yDataArray.setSize(totalSize+2);

	m_yDataArraySecondChannel = new float64[totalSize];
	memset(m_yDataArraySecondChannel, 0, totalSize * sizeof(float64)); //set to zero
	yDataArraySecondChannel.update(m_yDataArraySecondChannel);
	yDataArraySecondChannel.setSize(totalSize);
}


/*
 * Append single measure data from last spectrum measure to the acquisition file
 */
void CountWithSpectro::saveSpectroData() {
	int32 totalSize = YSIZE * numSpectroMeasures();
	yDataArray.update(m_yDataArray);
	yDataArray.setSize(totalSize);
}

/*
 * Append double measured data (switcher mode) from last spectrum measure to the acquisition file
 */
void CountWithSpectro::saveSwitcherSpectroData() {
	int32 totalSize = YSIZE * numSpectroMeasures();
	yDataArray.update(m_yDataArray);
	yDataArray.setSize(totalSize+2);

	yDataArraySecondChannel.update(m_yDataArraySecondChannel);
	yDataArraySecondChannel.setSize(totalSize+2);
}

/*
 * Update yDataArray property with the spectrum data corresponding to the current acquisition
 * This method assumes a shift of two positions between spectrum data and count data (due to use of SampleSequencer)
 * NO append done, spectrum data is saved in current count file
 */
void CountWithSpectro::saveSpectroDataInShiftedMode() {
	int32 size = spectro->size() * numSpectroMeasures();
	if (actualPosition() % 2) {

		// Save spectrum if there is any
		std::copy(m_yDataEvenPositionArray, m_yDataEvenPositionArray + size, m_yDataArrayToSave);

		// Reset temp array with last measure
		std::copy(m_yDataArray, m_yDataArray + size, m_yDataEvenPositionArray);

	} else {
		std::copy(m_yDataOddPositionArray, m_yDataOddPositionArray + size, m_yDataArrayToSave);

		// Reset temp array with last measure
		std::copy(m_yDataArray, m_yDataArray + size, m_yDataOddPositionArray);
	}

	int32 totalSize = YSIZE * numSpectroMeasures();
	yDataArray.update(m_yDataArrayToSave); // Change local array used to store propertyArray values
	yDataArray.setSize(totalSize);
}

float64 CountWithSpectro::getCountTimeInSec() {
	float64 time;
	if (timeType() == "h") {
		time = countTime() * 3600.;
	} else if (timeType() == "m") {
		time = countTime() * 60.;
	} else if (timeType() == "s") {
		time = countTime();
	}
	return time;
}

/**
 * This method calculates the waiting time necessary to distribute the spectrometer measures (in both single or switcher mode) equally through the acquisition time
 */
float64 CountWithSpectro::calculateWaitingTimeInSec() {
	//Convert all units into seconds
	float64 totalSpectroTime;
	float64 switcherSleep = (spectroSwitcher->SLEEP_TIME_MICROSEC) * 0.000001;//0.0001;
	float64 singleWaitTime = 0.0;

	if (numSpectroMeasures() > 1) {
		int32 scansToAverage = spectro->scansToAverage();

		if (useSwitcher()) {
			float64 measureTime = integrationTime() * 0.000001 + USB_DELAY_MILLISEC * 0.001;
			totalSpectroTime = 2 * (measureTime * numSpectroMeasures() * scansToAverage + switcherSleep);
		} else {
			float64 measureTime = integrationTime() * 0.000001 + USB_DELAY_MILLISEC * 0.001;
			totalSpectroTime = measureTime * numSpectroMeasures() * scansToAverage;
		}

		float64 countTime = getCountTimeInSec();
		float64 totalWaitTime = (countTime - totalSpectroTime);
		singleWaitTime = 0.95 * (totalWaitTime / (numSpectroMeasures() - 1)); //0.95 -> factor to avoid having spectro measures just at the end
	}
	return singleWaitTime;
}

/*
 * setEnabled
 */
void CountWithSpectro::setEnabledForLoading(bool all, bool postconf) {
	addExcludeSetEnableController(count.getRole());
	ExperimentController::setEnabledForLoading(all, postconf);
}

/*
 * setDisabled
 */
void CountWithSpectro::setDisabled(bool all) {
	addExcludeSetEnableController(count.getRole());
	ExperimentController::setDisabled(all);
}

void CountWithSpectro::cleanArrays() {
	if (m_yDataEvenPositionArray) {
		delete[] m_yDataEvenPositionArray;

	} else if (m_yDataOddPositionArray) {
		delete[] m_yDataOddPositionArray;

	} else if (m_yDataArrayToSave) {
		delete[] m_yDataArrayToSave;
	}

	int32 totalSize = xSize() * numSpectroMeasures();
	m_yDataEvenPositionArray = new float64[totalSize];
	memset(m_yDataEvenPositionArray, 0, totalSize * sizeof(float64)); //set to zero

	m_yDataOddPositionArray = new float64[totalSize];
	memset(m_yDataOddPositionArray, 0, totalSize * sizeof(float64)); //set to zero

	m_yDataArrayToSave = new float64[totalSize];
	memset(m_yDataArrayToSave, 0, totalSize * sizeof(float64)); //set to zero
}

}
