/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <boost/lexical_cast.hpp>
#include <common/base/Date.h>
#include <fstream>
#include <iostream>
#include "RemoteFileRep.h"

#include <thread>
#include <memory>

std::thread t;
std::unique_ptr<int> u;

namespace test {

using namespace std;
using namespace common;
using namespace boost;
using namespace cameo;

const std::string RemoteFileRep::TYPE = "remote_file_rep";
const std::string RemoteFileRep::REMOTE_APPLICATION = "nfilerep";

RemoteFileRep::RemoteFileRep(const std::string& name) :
	ExperimentController(name),
	controller::Stoppable(this) {

	serverEndpoint.init(this, SAVE, "cameo_server");
	initialized.init(this, NOSAVE, "initialized");
	filename.init(this, SAVE, "filename");
}

RemoteFileRep::RemoteFileRep(const RemoteFileRep& controller) :
	ExperimentController(controller),
	controller::Stoppable(this) {
}

RemoteFileRep::~RemoteFileRep() {

	// Stop the remote application.
	m_remoteApplication->stop();

	// Wait for the termination
	application::State state = m_remoteApplication->waitFor();

	cout << "Remote application " << m_remoteApplication->getName() << " terminated with state " << application::toString(state) << endl;
}

void RemoteFileRep::postConfiguration() {

	if (serverEndpoint() == "") {
		m_server.reset(new Server(application::This::getServer().getEndpoint()));

	} else {
		m_server.reset(new Server(serverEndpoint()));
	}

	initialized = initApplication();
}

bool RemoteFileRep::initApplication() {

	// Do not initialize if it is already done
	if (initialized()) {
		return true;
	}

	// Start the remote server
	if (!m_server->isAvailable(1000)) {
		cout << "Remote server is not available" << endl;
		return false;
	}

	cout << "Remote server is connected" << endl;

	m_remoteApplication = m_server->connect(REMOTE_APPLICATION);
	if (m_remoteApplication->exists()) {
		// The application exists from a previous server session
		m_remoteApplication->kill();
		application::State state = m_remoteApplication->waitFor();
		cout << "Terminated remote application with state " << application::toString(state) << endl;
	}

	m_remoteApplication = m_server->start(REMOTE_APPLICATION);

	if (!m_remoteApplication->exists()) {
		cout << "No remote application" << endl;
		return false;
	}

	// Application initialized
	initialized = true;

	return true;
}

void RemoteFileRep::sendFile() {

	if (!initApplication()) {
		return;
	}

	ifstream numorFile;
	numorFile.open(filename().c_str());

	stringstream stream;
	stream << numorFile.rdbuf();

	Date begin;

	string content = stream.str();

	// Create the requester
	auto_ptr<application::Requester> requester = application::Requester::create(*m_remoteApplication, "file_responder");

	if (requester.get() == 0) {
		cout << "requester error" << endl;
		return;
	}

	// Send the content in binary
	requester->sendBinary(content);

	string response;
	requester->receiveBinary(response);

	Date end;
	double ms = (end - begin).getMilliseconds();
	cout << "Sent and received file in " << ms << " ms (" << (content.size() / ms) << " ko/s)" << endl;
//	if (content.size() != result[0]) {
//		cout << "\tProblem with size" << endl;
//	}
}

void RemoteFileRep::start() {

	while (true) {
		// Send the numor
		sendFile();

		if (isStopped()) {
			break;
		}
	}
}

void RemoteFileRep::stop() {

}

}
