/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "RemoteDPPLiveProcess.h"
#include <common/base/ServerProperties.h>

using boost::property_tree::ptree;

namespace npp {

using namespace std;
using namespace common;
using namespace boost;
using namespace cameo;
using namespace lstdpp128;

const std::string RemoteDPPLiveProcess::TYPE = "remote_dpp_live_process";
const std::string RemoteDPPLiveProcess::TEST_MODE = "test";
const std::string RemoteDPPLiveProcess::FILE_MODE = "file";
const std::string RemoteDPPLiveProcess::DIRECT_MODE = "direct";

RemoteDPPLiveProcess::RemoteDPPLiveProcess(const std::string& name) :
	DPPLiveProcess(name) {

	serverEndpoint.init(this, SAVE, "cameo_server");
	blockSize.init(this, SAVE, "block_size");
	mode.init(this, SAVE, "mode");
	testFileName.init(this, SAVE, "test_file_name");
	refreshTimeS.init(this, SAVE, "refresh_time_s");
	sleepTimeMS.init(this, SAVE, "sleep_time_ms");
}

RemoteDPPLiveProcess::RemoteDPPLiveProcess(const RemoteDPPLiveProcess& controller) :
	DPPLiveProcess(controller) {
}

RemoteDPPLiveProcess::~RemoteDPPLiveProcess() {
}

void RemoteDPPLiveProcess::postConfiguration() {

	if (serverEndpoint() == "") {
		m_server.reset(new Server(application::This::getServer().getEndpoint()));

	} else {
		m_server.reset(new Server(serverEndpoint()));
	}
}

bool RemoteDPPLiveProcess::fileReadingMode() {

	if (mode() == TEST_MODE || mode() == FILE_MODE) {
		return true;
	}

	// The direct mode is not supported now.

	return false;
}

std::string RemoteDPPLiveProcess::getFileNameToRead() {

	if (mode() == TEST_MODE) {
		return testFileName();
	}
	else {
		// Set the numor name.
		string numorName = fileName();
		size_t size = numorName.size();
		if (size < 6) {
			string prefix(6 - size, '0');
			numorName = prefix + numorName;
		}

		return ServerProperties::getInstance()->getNomadDataPath() + "/" + numorName + ".lst";
	}

	return "";
}

void RemoteDPPLiveProcess::initFileReading() {

	if (fileReadingMode()) {
		m_reader.reset(new lstdpp128::LiveReader(blockSize()));
		m_reader->open(getFileNameToRead());
	}
}

void RemoteDPPLiveProcess::fileReadingLoop() {

	if (fileReadingMode()) {

		cout << "reading file " << getFileNameToRead() << endl;

		// Loop over events
		while (!isStopped()) {

			// Reading next block
			int32_t blockSize = m_reader->read();

			// Exiting if no data
			if (blockSize == 0) {
				break;
			}

			// Sending block
			int32_t * memoryBlock = m_reader->buffer();

			// Put the array into a string
			string data(reinterpret_cast<char *>(memoryBlock), sizeof(int32_t) * blockSize);

			// Send the data
			m_publisher->sendBinary(data);

			// Sleep for milliseconds.
			boost::this_thread::sleep(boost::posix_time::milliseconds(sleepTimeMS()));
		}
		cout << "controller has stopped" << endl;
		
		m_publisher->sendEnd();
	
		cout << "finished reading file" << endl;
	}
		
}

std::auto_ptr<boost::thread> RemoteDPPLiveProcess::startFileReading() {

	if (fileReadingMode()) {
		return auto_ptr<thread>(new thread(bind(&RemoteDPPLiveProcess::fileReadingLoop, this)));
	}
	return auto_ptr<thread>();
}

void RemoteDPPLiveProcess::finishFileReading(std::auto_ptr<boost::thread> thread) {

	if (fileReadingMode()) {
		// Join the thread
		thread->join();
	}
}

void RemoteDPPLiveProcess::stop() {

	// Stop the reader which should be blocked in fileReadingLoop.
	if (m_reader.get() != 0) {
		m_reader->stop();
	}
}

boost::property_tree::ptree RemoteDPPLiveProcess::writeListModeContextToPropertyTree(const ListModeContext & listModeContext) {

	ptree root;

	root.put("version", listModeContext.version);
	root.put("timeBase", listModeContext.timeBase);
	root.put("nbEvents", listModeContext.nbEvents);

	ptree crateBoard;

	for (int c = 0; c < listModeContext.crateBoard.nbCrates; c++) {

		ptree crate;

		for (int b = 0; b < listModeContext.crateBoard.crates[c].nbBoards; b++) {

			ptree board;

			board.put("eventType", listModeContext.crateBoard.crates[c].boards[b].eventType);
			board.put("crate", listModeContext.crateBoard.crates[c].boards[b].crate);
			board.put("nbChannels", listModeContext.crateBoard.crates[c].boards[b].nbChannels);
			board.put("boardType", listModeContext.crateBoard.crates[c].boards[b].boardType);

			crate.push_back(make_pair("", board));
		}

		crateBoard.push_back(make_pair("", crate));
	}

	root.put_child("crateboard", crateBoard);

	return root;
}

}

