/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "QE65000Controller.h"
#include "controllers/common/baselegacy/ExperimentControllerDef.h"
#include "controllers/common/family/Families.h"
#include <common/base/Date.h>
using namespace std;

namespace spectrometer {

const string QE65000Controller::TYPE = "spectro";

const std::string QE65000Controller::IDLE = "idle";
const std::string QE65000Controller::SEQUENCE = "sequence";
const std::string QE65000Controller::COUNT_SINGLE = "count_single";
const std::string QE65000Controller::COUNT_SWITCHER = "count_switcher";

QE65000Controller::QE65000Controller(const std::string& name) :
		ExperimentController(TYPE, name), controller::Stoppable(this) {

	setFamily(family::SAMPLE_ENVIRONMENT);

	// force register function to allow the controller to be used by the "classic" UVspectrometer controller.
	registerFunction(SHUTTER_EXPERIMENT_CONTROLLER);

	integrationTime.init(this, SAVE, "integration_time");
	xData.init(this, NOSAVE, "x_data");
	yData.init(this, NOSAVE, "y_data");
	size.init(this, SAVE, "size");
	spectroStatus.init(this, SAVE, "spectro_status");
	spectroStatusMessage.init(this, SAVE, "spectro_status_message");
	spectroOpenErrorMessage.init(this, SAVE, "spectro_open_error_message");
	mode.init(this, SAVE, "mode");
	spectroMode.init(this, SAVE, "spectro_mode");
	plotName.init(this, SAVE, "plot_name");
	numor.init(this, SAVE, "numor");
	scansToAverage.init(this, SAVE, "scans_average");

	registerFunction(TYPE);

	spectroDriver.init(this, "driver");

	m_StopActivated = false;
}

QE65000Controller::~QE65000Controller() {
}

void QE65000Controller::postConfiguration() {
	//Pass info to the driver
	registerRefresher(integrationTime, &QE65000Controller::refreshIntegrationTime, this);

	//Get info from driver
	registerUpdater(spectroDriver->status, &QE65000Controller::updateStatus, this);
	registerUpdater(spectroDriver->openError, &QE65000Controller::updateOpenError, this);
	registerUpdater(spectroDriver->size, &QE65000Controller::updateSize, this);

	registerProgression(spectroDriver, &QE65000Controller::updateProgression, this);

	//Set initial property values
	spectroStatusMessage.update("UNPLUGGED");

	mode.update(IDLE);
	//plotName.update(""); //original value = empty

	//Get size from driver
	size = spectroDriver->size();
}

void QE65000Controller::refreshIntegrationTime(int32 time) {
	spectroDriver->integrationTime = time;
}

void QE65000Controller::updateStatus() {
	spectroStatus.update(spectroDriver->status());
	if (spectroStatus() == 0) {
		spectroStatusMessage.update("UNPLUGGED");
	} else {
		spectroStatusMessage.update("PLUGGED");
	}
}

void QE65000Controller::updateOpenError() {
	if (spectroDriver->openError()) {
		spectroOpenErrorMessage.update("USB OPEN ERROR");
	} else {
		spectroOpenErrorMessage.update("");
	}
}

void QE65000Controller::updateSize() {
	size.update(spectroDriver->size());
}

void QE65000Controller::updateProgression() {
	int32 progression = 0;
	commandProgression = progression;
}

void QE65000Controller::start() {
	commandProgression = 0;
	m_StopActivated = false;

	if (mode() != SEQUENCE) {
		numor.update(0);
	}

	//Get size from driver
	size = spectroDriver->size();

	if (isConnected() && (!spectroDriver->openError())) {

		common::Date startTime = common::Date();

		//Take into account number of scans to average
		if (scansToAverage() == 1) {
			commandProgression = 50;
			spectroDriver.execute("start", false);
			commandProgression = 75;

			//x data
			xData.update(spectroDriver->xData());
			xData.setSize(size());

			//y data
			yData.update(spectroDriver->yData());
			yData.setSize(size());

			//To take those changes into account do the following:
			xData.sendEvent();
			yData.sendEvent();

			commandProgression = 99;
			spectroDriver.execute("stop", false);

		} else {
			//REVISE THE USAGE OF THE ARRAYS HERE
			float64* tempYData;
			float64* finalYData;

			for (int32 i = 0; i < scansToAverage(); i++) {

				if (!m_StopActivated) {
					commandProgression = i * 100 / scansToAverage();

					spectroDriver.execute("start", false);

					//During first scan get size of array and initialize everything
					if (i == 0) {
						tempYData = new float64[size()];
						finalYData = new float64[size()];

						//to avoid problems with incorrect initial values initialize content of both arrays to zero
						for (int32 j = 0; j < size(); j++) {
							tempYData[j] = 0;
							finalYData[j] = 0;
						}
					}

					//y data
					tempYData = spectroDriver->yData();
					for (int32 j = 0; j < size(); j++) {
						finalYData[j] = finalYData[j] + tempYData[j];
						if (i == scansToAverage() - 1) {
							finalYData[j] = finalYData[j] / scansToAverage();
						}
					}

				}
			}

			if (!m_StopActivated) {
				xData.update(spectroDriver->xData());
				xData.setSize(size());

				yData.update(finalYData);
				yData.setSize(size());

				//To take those changes into account do the following:
				xData.sendEvent();
				yData.sendEvent();
			}

			commandProgression = 99;
			spectroDriver.execute("stop", false);
		}

		//For debug purposes
		common::Date endTime = common::Date();
		common::Duration duration = endTime - startTime;
		cout << "TOTAL TIME " << duration.getMilliseconds() << endl;
	}

}

void QE65000Controller::stop() {
	spectroDriver.execute("stop", false);
	m_StopActivated = true;
	commandProgression = 100;
}

bool QE65000Controller::isConnected() {
	return (spectroDriver->status() != 0);
}

void QE65000Controller::setPlotName(int32 plotId) {
	numor.update(plotId);
}

}

