/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "VEXPController.h"
#include "controllers/common/family/Families.h"
#include "contrib/rapidjson/document.h"
#include "contrib/rapidjson/reader.h"
#include "contrib/rapidjson/writer.h"
#include "contrib/rapidjson/stringbuffer.h"
#include "contrib/rapidjson/prettywriter.h"

using namespace boost;
using namespace std;

namespace vexp {

const string VEXPController::TYPE = "vexp_controller";

VEXPController::VEXPController(const string& name) :
	ExperimentController(name) {

	setFamily(family::HIDDEN);

	as.init(this, NOSAVE, "as");
	bs.init(this, NOSAVE, "bs");
	cs.init(this, NOSAVE, "cs");
	aa.init(this, NOSAVE, "aa");
	bb.init(this, NOSAVE, "bb");
	cc.init(this, NOSAVE, "cc");
	ax.init(this, NOSAVE, "ax");
	ay.init(this, NOSAVE, "ay");
	az.init(this, NOSAVE, "az");
	bx.init(this, NOSAVE, "bx");
	by.init(this, NOSAVE, "by");
	bz.init(this, NOSAVE, "bz");
	u.init(this, NOSAVE, "Umatrix");
	b.init(this, NOSAVE, "Bmatrix");

	qh.init(this, NOSAVE, "qh");
	qk.init(this, NOSAVE, "qk");
	ql.init(this, NOSAVE, "ql");
	en.init(this, NOSAVE, "en");
	qm.init(this, NOSAVE, "qm");

	ki.init(this, NOSAVE, "ki");
	kf.init(this, NOSAVE, "kf");

	a1.init(this, NOSAVE, "a1");
	a2.init(this, NOSAVE, "a2");
	a3.init(this, NOSAVE, "a3");
	a4.init(this, NOSAVE, "a4");
	a5.init(this, NOSAVE, "a5");
	a6.init(this, NOSAVE, "a6");


	calcqh.init(this, NOSAVE, "calcqh");
	calcqk.init(this, NOSAVE, "calcqk");
	calcql.init(this, NOSAVE, "calcql");
	calcen.init(this, NOSAVE, "calcen");
	calcqm.init(this, NOSAVE, "calcqm");

	calcki.init(this, NOSAVE, "calcki");
	calckf.init(this, NOSAVE, "calckf");

	calcvalues.init(this, NOSAVE, "calcvalues");

	calca1.init(this, NOSAVE, "calca1");
	calca2.init(this, NOSAVE, "calca2");
	calca3.init(this, NOSAVE, "calca3");
	calca4.init(this, NOSAVE, "calca4");
	calca5.init(this, NOSAVE, "calca5");
	calca6.init(this, NOSAVE, "calca6");

	calctrigger.init(this, NOSAVE, "calctrigger");


	calculationStatus.init(this, NOSAVE, "calculation_status");
	errorMessage.init(this, NOSAVE, "error_message");
	moveModel.init(this, SAVE, "move_model");

	tasSettings.init(this, "tas_settings");
	sample.init(this, "sample");
	scattering.init(this, "scattering");
	incidentBeam.init(this, "incident_beam");
	scatteredBeam.init(this, "scattered_beam");

	a1Controller.init(this, "A1");
	a2Controller.init(this, "A2");
	a3Controller.init(this, "A3");
	a4Controller.init(this, "A4");
	a5Controller.init(this, "A5");
	a6Controller.init(this, "A6");
}

VEXPController::VEXPController(const VEXPController& controller) :
	ExperimentController(controller) {
}

VEXPController::~VEXPController() {
}

void VEXPController::updateUMatrix() {
	u.update(sample->uRef());
	u.setSize(sample->uRef.getSize());
	u.sendEvent();
}

void VEXPController::updateBMatrix() {
	b.update(tasSettings->b());
	b.setSize(tasSettings->b.getSize());
	b.sendEvent();
}

void VEXPController::refreshCalcParameters(const std::string& value) {

	cout << "refreshCalcParameters " << value << endl;

	// Parse the values.
	rapidjson::Document document;
	document.Parse(value.c_str());

	rapidjson::Value& ki = document["ki"];
	rapidjson::Value& kf = document["kf"];
	rapidjson::Value& qh = document["qh"];
	rapidjson::Value& qk = document["qk"];
	rapidjson::Value& ql = document["ql"];

	// Calculate the angles and the status.
	calculate(ki.GetDouble(), kf.GetDouble(), qh.GetDouble(), qk.GetDouble(), ql.GetDouble());
}

void VEXPController::refreshCalcTrigger(int32 value) {

	cout << "refreshCalcTrigger " << value << endl;

	rapidjson::Document document;

	rapidjson::StringBuffer s;
	rapidjson::PrettyWriter<rapidjson::StringBuffer> writer(s);
	s.Clear();
	writer.StartObject();

	writer.Key("ki");
	writer.Double(calcki());

	writer.Key("kf");
	writer.Double(calckf());

	writer.Key("qh");
	writer.Double(calcqh());

	writer.Key("qk");
	writer.Double(calcqk());

	writer.Key("ql");
	writer.Double(calcql());

	writer.EndObject();

	cout << "JSON parameters = " << s.GetString() << endl;
	//calculate(calcki(), calckf(), calcqh(), calcqk(), calcql());

	refreshCalcParameters(s.GetString());
}

void VEXPController::refreshFloat64Property(SimpleProperty<float64>& property, float64 value) {

	// Update the value.
	property = value;
}

void VEXPController::refreshFloat64PropertyToMoveModel(SimpleProperty<float64>& property, float64 value) {

	// Update the value only in case moveModel is true.
	if (moveModel()) {
		property = value;
	}
}

void VEXPController::updateProperties() {

	// Copy the values.
	as.update(tasSettings->as());
	bs.update(tasSettings->bs());
	cs.update(tasSettings->cs());
	aa.update(tasSettings->aa());
	bb.update(tasSettings->bb());
	cc.update(tasSettings->cc());
	ax.update(tasSettings->ax());
	ay.update(tasSettings->ay());
	az.update(tasSettings->az());
	bx.update(tasSettings->bx());
	by.update(tasSettings->by());
	bz.update(tasSettings->bz());

	updateUMatrix();
	updateBMatrix();

	qh.update(scattering->qh());
	qk.update(scattering->qk());
	ql.update(scattering->ql());
	en.update(scattering->en());
	qm.update(scattering->qm());

	ki.update(incidentBeam->ki());
	kf.update(scatteredBeam->kf());

	a1.update(a1Controller->position());
	a2.update(a2Controller->position());
	a3.update(a3Controller->position());
	a4.update(a4Controller->position());
	a5.update(a5Controller->position());
	a6.update(a6Controller->position());
}

void VEXPController::postConfiguration() {

	// Register the property updaters.
	registerPropertyCopierByUpdate(tasSettings->as, as);
	registerPropertyCopierByUpdate(tasSettings->bs, bs);
	registerPropertyCopierByUpdate(tasSettings->cs, cs);
	registerPropertyCopierByUpdate(tasSettings->aa, aa);
	registerPropertyCopierByUpdate(tasSettings->bb, bb);
	registerPropertyCopierByUpdate(tasSettings->cc, cc);
	registerPropertyCopierByUpdate(tasSettings->ax, ax);
	registerPropertyCopierByUpdate(tasSettings->ay, ay);
	registerPropertyCopierByUpdate(tasSettings->az, az);
	registerPropertyCopierByUpdate(tasSettings->bx, bx);
	registerPropertyCopierByUpdate(tasSettings->by, by);
	registerPropertyCopierByUpdate(tasSettings->bz, bz);

	registerUpdater(sample->uRef, &VEXPController::updateUMatrix, this);
	registerUpdater(sample->b, &VEXPController::updateBMatrix, this);

	registerPropertyCopierByUpdate(scattering->qh, qh);
	registerPropertyCopierByUpdate(scattering->qk, qk);
	registerPropertyCopierByUpdate(scattering->ql, ql);
	registerPropertyCopierByUpdate(scattering->en, en);
	registerPropertyCopierByUpdate(scattering->qm, qm);

	registerPropertyCopierByUpdate(incidentBeam->ki, ki);
	registerPropertyCopierByUpdate(scatteredBeam->kf, kf);

	registerPropertyCopierByUpdate(a1Controller->position, a1);
	registerPropertyCopierByUpdate(a2Controller->position, a2);
	registerPropertyCopierByUpdate(a3Controller->position, a3);
	registerPropertyCopierByUpdate(a4Controller->position, a4);
	registerPropertyCopierByUpdate(a5Controller->position, a5);
	registerPropertyCopierByUpdate(a6Controller->position, a6);


	// To calculate the angles.
	registerRefresher(calctrigger, &VEXPController::refreshCalcTrigger, this);
	registerRefresher(calcvalues, &VEXPController::refreshCalcParameters, this);

	// To move the model.
	registerRefresher(a1, &VEXPController::refreshFloat64PropertyToMoveModel, this, a1Controller->position);
	registerRefresher(a2, &VEXPController::refreshFloat64PropertyToMoveModel, this, a2Controller->position);
	registerRefresher(a3, &VEXPController::refreshFloat64PropertyToMoveModel, this, a3Controller->position);
	registerRefresher(a4, &VEXPController::refreshFloat64PropertyToMoveModel, this, a4Controller->position);
	registerRefresher(a5, &VEXPController::refreshFloat64PropertyToMoveModel, this, a5Controller->position);
	registerRefresher(a6, &VEXPController::refreshFloat64PropertyToMoveModel, this, a6Controller->position);

	updateProperties();
}

void VEXPController::calculate(float64 ki, float64 kf, float64 qh, float64 qk, float64 ql) {

	// Reset status.
	calculationStatus.setIdle();

	// Ki.
	incidentBeam->calcki = ki;

	// calca1, calca2, calcStatus are calculated.
	incidentBeam->calculate(tas::Beam::K_MODE);

	calca1 = incidentBeam->calca1();
	calca2 = incidentBeam->calca2();

	cout << "calc A1 " << calca1() << endl;
	cout << "calc A2 " << calca2() << endl;
	cout << "calc A3 " << calca3() << endl;
	cout << "calc A4 " << calca4() << endl;
	cout << "calc A5 " << calca5() << endl;
	cout << "calc A6 " << calca6() << endl << endl;

	// Manage the status.
	if (incidentBeam->calcStatus.isError()) {
		calculationStatus.setError();
	}
	else if (incidentBeam->calcStatus.isWarning()) {
		calculationStatus.setWarning();
	}

	//cout << "Ki -> status " << incidentBeam->calcStatus() << endl;

	// Kf.
	scatteredBeam->calckf = kf;

	// calca1, calca2, calcStatus are calculated.
	scatteredBeam->calculate(tas::Beam::K_MODE);

	calca5 = scatteredBeam->calca5();
	calca6 = scatteredBeam->calca6();

	cout << "calc A1 " << calca1() << endl;
	cout << "calc A2 " << calca2() << endl;
	cout << "calc A3 " << calca3() << endl;
	cout << "calc A4 " << calca4() << endl;
	cout << "calc A5 " << calca5() << endl;
	cout << "calc A6 " << calca6() << endl << endl;

	// Manage the status.
	if (scatteredBeam->calcStatus.isError()) {
		calculationStatus.setError();
	}
	else if (scatteredBeam->calcStatus.isWarning() && !calculationStatus.isError()) {
		calculationStatus.setWarning();
	}

	//cout << "Kf -> status " << scatteredBeam->calcStatus() << endl;

	// Q.
	scattering->calcqh = qh;
	scattering->calcqk = qk;
	scattering->calcql = ql;
	scattering->calcen = 2.072 * (ki * ki - kf * kf);
	scattering->calcki = ki;
	scattering->calckf = kf;

	// Calculate by fixing Ki.
	scattering->calculate(tas::Beam::K_MODE, 1);

	calca3 = scattering->calca3();
	calca4 = scattering->calca4();

	cout << "calc A1 " << calca1() << endl;
	cout << "calc A2 " << calca2() << endl;
	cout << "calc A3 " << calca3() << endl;
	cout << "calc A4 " << calca4() << endl;
	cout << "calc A5 " << calca5() << endl;
	cout << "calc A6 " << calca6() << endl << endl;

	// Manage the status.
	if (scattering->calcStatus.isError()) {
		calculationStatus.setError();
	}
	else if (scattering->calcStatus.isWarning() && !calculationStatus.isError()) {
		calculationStatus.setWarning();
		cerr << "Warning: Calculated Kf are different" << endl;
	}

	//cout << "Q -> status " << scattering->calcStatus() << endl;
}

}
