/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "RealCAENCfdDriver.h"
#include "CAENCfdDriver.h"
#include <controllers/common/acquisition/mode/DPPAcquisitionCommon.h>

using namespace std;

namespace caen_cfd {

/*
 * Constructor
 */
RealCAENCfdDriver::RealCAENCfdDriver(CAENCfdDriver* owner) : CAENCfdState(owner) {

	m_boardHandle = 0;
	m_boardInit = 0;
	ret = CAEN_DGTZ_Success;
}

/*
 * Destructor
 */
RealCAENCfdDriver::~RealCAENCfdDriver() {

}

/*!
 * \brief open command implementation
 */
void RealCAENCfdDriver::open() {

	boardParams.LinkType = (CAEN_DGTZ_ConnectionType)owner()->linkType();
	boardParams.VMEBaseAddress = owner()->address();
	ret = CAEN_DGTZ_OpenDigitizer(boardParams.LinkType, 0, 0, boardParams.VMEBaseAddress, &m_boardHandle);
	if (ret) {
		cerr << "Error CAEN_DGTZ_OpenDigitizer: " << ret << endl;
		return;
	}

}

/*!
 * \brief close command implementation
 */
void RealCAENCfdDriver::close() {

	CAEN_DGTZ_CloseDigitizer(m_boardHandle);
}

/*!
 * \brief Init command implementation
 */
void RealCAENCfdDriver::init() {

	reset();

	readInfos();

	open();

	ret = 0;
	// Enable channels
	boardParams.ChannelMask = owner()->channelMask();
	ret = CAEN_DGTZ_SetChannelEnableMask(m_boardHandle, boardParams.ChannelMask);
	if (ret)
		cerr << "Error CAEN_DGTZ_SetChannelEnableMask: " << ret << endl;

	close();

	// Set acquisition mode
	setAcqMode();

	open();

	// Set the I/O level (NIM or TTL)
	boardParams.IOLevel = (CAEN_DGTZ_IOLevel_t)owner()->ioLevel();
	ret |= CAEN_DGTZ_SetIOLevel(m_boardHandle, boardParams.IOLevel);
	if (ret)
		cerr << "Error CAEN_DGTZ_SetIOLevel: " << ret << endl;

	// Set start acquisition mode
	uint32_t d32 = 0;
	ret |= CAEN_DGTZ_ReadRegister(m_boardHandle, CAEN_DGTZ_ACQ_CONTROL_ADD, &d32);
	// Arm acquisition (Run will start when SIN goes high)
	ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, CAEN_DGTZ_ACQ_CONTROL_ADD, (d32 & 0xFFFFFFF0) | 0xD);
	if (ret)
		cerr << "Error setting start acquisition mode: " << ret << endl;

	// Disable trig_in mode
	ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, CAEN_DGTZ_TRIGGER_SRC_ENABLE_ADD, 0x80000000);
	if (ret)
		cerr << "Error disabling trig_in: " << ret << endl;

	// Disable trig_out mode
	ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8110, 0);
	if (ret)
		cerr << "Error disabling trig_out: " << ret << endl;

	// Buffer organization
	ret |= CAEN_DGTZ_SetDPPEventAggregation(m_boardHandle, boardParams.EventAggr, 0);
	if (ret)
		cerr << "Error CAEN_DGTZ_SetDPPEventAggregation: " << ret << endl;

	// Number of buffers per BLT
	ret |= CAEN_DGTZ_SetMaxNumAggregatesBLT(m_boardHandle, 1023);
	if (ret)
		cerr << "Error CAEN_DGTZ_SetMaxNumAggregatesBLT: " << ret << endl;

	close();

	writeParam();

	open();

	// Set virtual probe settings
	if (owner()->modelName() == "V1724") {
		ret |= CAEN_DGTZ_SetDPP_PHA_VirtualProbe(m_boardHandle, CAEN_DGTZ_DPP_VIRTUALPROBE_DUAL,
				CAEN_DGTZ_DPP_PHA_VIRTUALPROBE1_trapezoid, CAEN_DGTZ_DPP_PHA_VIRTUALPROBE2_Input,
				CAEN_DGTZ_DPP_PHA_DIGITAL_PROBE_Peaking);
	} else {
		ret |= CAEN_DGTZ_SetDPP_VirtualProbe(m_boardHandle, ANALOG_TRACE_1, CAEN_DGTZ_DPP_PHA_VIRTUALPROBE1_trapezoid);
		ret |= CAEN_DGTZ_SetDPP_VirtualProbe(m_boardHandle, ANALOG_TRACE_2, CAEN_DGTZ_DPP_PHA_VIRTUALPROBE2_Input);
		ret |= CAEN_DGTZ_SetDPP_VirtualProbe(m_boardHandle, DIGITAL_TRACE_1, CAEN_DGTZ_DPP_DIGITALPROBE_Peaking);
	}
	if (ret)
		cerr << "Error CAEN_DGTZ_SetDPP_VirtualProbe: " << ret << endl;

	// Calibrate ADC channels at startup
	if (m_boardInit == 0) {
		uint32_t d32 = 0;
		cout << "INFO: Calibrating ADCs... " << endl;
		ret |= CAEN_DGTZ_Calibrate(m_boardHandle);
		if (ret)
			cerr << "Error CAEN_DGTZ_Calibrate: " << ret << endl;
		for (int32 i=0; i<owner()->channels(); ++i) {
			LockTempCalibration_x730(i);
			for (int32 i=0; i<owner()->channels(); ++i) {
			CAEN_DGTZ_ReadRegister(m_boardHandle, 0x1088 + (i<<8), &d32);
			if (!(d32 & 0x8))
				cout << "WARNING: calibration not done on ch. " << i << endl;
			}
		}
		usleep(1000);
		m_boardInit = 1;
	}

/*
	// PN1 COINCIDENCE MODE =====> HACK EMILIO 19/01/2018
	if (owner()->usePn1Coincidence() == 1) {
		// Channel CTRL register (individual trigger, sequencial readout)
		ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8000, 0x010E0115);

		if (ret) {
			cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
			close();
			return;
		}
	} else {
		// Set acquisition mode
		if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::LISTMODE_MODE) {
			ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8000, 0x010E0115);

			if (ret) {
				cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
				close();
				return;
			}

		} else if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::SCOPE_MODE) {
			if (owner()->cfdTrap() == 1) {
				ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8000, 0x014F2914);
			} else {
				ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8000, 0x014F3914);
			}

			if (ret) {
				cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
				close();
				return;
			}

		} else { // WAVEFORM NO DUAL TRACE
			ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8000, 0x010F0115);

			if (ret) {
				cerr << "Error CAEN_DGTZ_WriteRegister: " << ret << endl;
				close();
				return;
			}

		}
	}
*/
	close();

	if (ret)
		cout << "******** WARNING ADC not initialized properly! ********" << endl;
}

/*!
 * \brief Clear command implementation
 */
void RealCAENCfdDriver::clear() {

	open();
	ret = 0;
	ret |= CAEN_DGTZ_ClearData(m_boardHandle);
	if (ret)
		cerr << "Error CAEN_DGTZ_ClearData: " << ret << endl;
	close();
}

/*!
 * \brief Reset command implementation
 */
void RealCAENCfdDriver::reset() {

	open();
	ret = 0;
	ret = CAEN_DGTZ_Reset(m_boardHandle);
	if (ret)
		cerr << "Error CAEN_DGTZ_Reset: " << ret << endl;
	close();
}

/*!
 * \brief Write Param command implementation
 */
void RealCAENCfdDriver::writeParam() {

	open();

	for (int32 i=0; i<owner()->channels(); ++i) {

		if (owner()->channelMask() & (1 << i)) {
			// Set DC offset
			int32 offset = owner()->dcOffset.get(i);
			CAEN_DGTZ_PulsePolarity_t polarity = (CAEN_DGTZ_PulsePolarity_t)owner()->pulsePolarity.get(i);
			uint16_t offs;
			if (polarity == 1) { 										// negative
				offs = (uint16_t)((offset * 65535) / 100);
			} else {  													// positive
				offs = (uint16_t)(((100 - offset) * 65535) / 100);
			}
			ret |= CAEN_DGTZ_SetChannelDCOffset(m_boardHandle, i, offs);
			if (ret)
				cerr << "Error CAEN_DGTZ_SetChannelDCOffset" << endl;

			ret |= CAEN_DGTZ_SetDPPPreTriggerSize(m_boardHandle, i,owner()->preTrigger.get(i));
			if (ret)
				cerr << "Error CAEN_DGTZ_SetDPPPreTriggerSize" << endl;

			int32_t MM, m, k, b, a, shf, ftd, thr, trgho, tu, tnorm;
			float Tg;

			int32_t tSampling = 4;
			tu = (tSampling * 4) * (1 << owner()->decimation.get(i));
			MM =  (int32_t)(owner()->signalDecayTime.get(i) / tu);
			k =   (int32_t)(owner()->trapezoidRiseTime.get(i) / tu);
			m =   (int32_t)(owner()->trapezoidFlatTop.get(i) / tu);
			ftd = (int32_t)(owner()->flatTopDelay.get(i) / tu);
			if (ftd > m) {
				cout << "WARNING: PeakingTime > FlatTop; forced PeakingTime = FlatTop" << endl;
				ftd = m;
			}
			b = (int32_t)( owner()->signalRiseTime.get(i) / (tSampling * 4));  // NOTE: timing parameters of the RCCR2 filters are not affected by the decimation
			a = (int32_t)( owner()->triggerSmoothing.get(i));
			thr = (int32_t)(owner()->triggerThreshold.get(i));
			trgho = (int32_t)(owner()->triggerHoldOff.get(i) / (tSampling * 4));
			Tg = (MM * k) / (owner()->trapezoidGain.get(i));

			// Set trapezoid gain and normalization factor
			for (shf=0; shf<32; shf++) {
				if ((1<<shf) > (int32_t)(Tg))
				break;
			}
			shf--;

			// Set control register to 0
			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1080 + (i<<8), 0);

			// Normalize trapezoid height
			tnorm = (uint32_t)(65535.0 * (1<<shf) / Tg);
			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x10C4 + (i<<8), tnorm);

			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1038 + (i<<8), owner()->preTrigger.get(i) / 4);			// pretrg

			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1054 + (i<<8), a);		// a

			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1058 + (i<<8), b);		// b

			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x105C + (i<<8), k);		// k

			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1060 + (i<<8), m);		// m

			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1064 + (i<<8), ftd);	// ftd

			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1068 + (i<<8), MM);		// MM

			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x106C + (i<<8), thr);  	// thr

			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1070 + (i<<8), 0);		//tww

			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1074 + (i<<8), trgho);	// trgho

			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1078 + (i<<8), 8);		// pkho

			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x107C + (i<<8), 4);		// blho

			registerSetBits(0x1080 + (i<<8), 0, 5, shf);														// shf

			registerSetBits(0x1080 + (i<<8), 12, 13, int32(owner()->peakMean.get(i)));							// nspk

			registerSetBits(0x1080 + (i<<8), 16, 16, int32(owner()->pulsePolarity.get(i)));						// polarity

			registerSetBits(0x1080 + (i<<8), 20, 22, int32(owner()->baselineMean.get(i)));						// nsbl

			registerSetBits(0x1080 + (i<<8), 8, 9, int32(owner()->decimation.get(i)));							// decimation

			registerSetBits(0x1080 + (i<<8), 26, 26, 1); 		// enable rollover tracing												// enable rollover tracing

			// Write CFD parameters
			setCfdParam();

			// Dynamic range: 0 = 2V; 1 = 0.5V
			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1028 + (i<<8), owner()->dynamicRange.get(i) & 1);  	// dynamic range 0=2Vpp, 1=0.5Vpp

			if (ret)
				cerr << "Error writeParam" << endl;
		}
	}

	close();
}

/*!
 * \brief Set acquisition mode command implementation
 */
void RealCAENCfdDriver::setAcqMode() {

	if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::LISTMODE_MODE) {
		boardParams.DPPAcqMode = CAEN_DGTZ_DPP_ACQ_MODE_List;
		boardParams.EventAggr = 0;
	} else if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::SCOPE_MODE) {
		boardParams.DPPAcqMode = CAEN_DGTZ_DPP_ACQ_MODE_Mixed;
		boardParams.EventAggr = 1;
	} else if (owner()->dppAcqMode() == acquisition::DPPAcquisitionCommon::WAVEFORM_MODE) {
		boardParams.DPPAcqMode = CAEN_DGTZ_DPP_ACQ_MODE_Mixed;
		boardParams.EventAggr = 1;
	}

	ret = 0;

	open();

	// Acquisition mode (LIST, OSCILLOSCOPE or MIXED)
	ret |= CAEN_DGTZ_SetDPPAcquisitionMode(m_boardHandle, boardParams.DPPAcqMode, CAEN_DGTZ_DPP_SAVE_PARAM_EnergyAndTime);
	ret |=CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8004, 1<<17); 		// Enable Extra Word
	// some specific settings in the global CTRL register
	ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8004, (1<<2));   	// set trgout bidir
	ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8004, 1);  			// Enable auto-flush

	// Set the number of samples in the waveform (same for all channels)
	// Set nb of samples for oscilloscope mode
	boardParams.RecordLength = owner()->recordLength();
	ret |= CAEN_DGTZ_SetRecordLength(m_boardHandle, boardParams.RecordLength, -1);
	// Library does not support length > 32K
	if (boardParams.RecordLength > 30000)
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8020, boardParams.RecordLength / 8);
	if (ret)
		cerr << "Error CAEN_DGTZ_SetDPPAcquisitionMode: " << ret << endl;

	close();
}

/*!
 * \brief Read Infos command implementation
 */
void RealCAENCfdDriver::readInfos() {

	open();
	ret = CAEN_DGTZ_GetInfo(m_boardHandle, &boardInfo);
	close();
	owner()->nbBits = (uint32_t)boardInfo.ADC_NBits;
	owner()->model = boardInfo.Model;
	owner()->channels = boardInfo.Channels;
	owner()->serialNumber = boardInfo.SerialNumber;
	owner()->modelName = boardInfo.ModelName;
	owner()->rocFirmware = boardInfo.ROC_FirmwareRel;
	owner()->amcFirmware = boardInfo.AMC_FirmwareRel;
}

/*!
 * \brief Read Status command implementation
 */
void RealCAENCfdDriver::readStatus() {

	open();

	uint32_t val;
	for (int32 i=0; i<owner()->channels(); ++i) {
		ret |= CAEN_DGTZ_ReadRegister(m_boardHandle, 0x10A8 + (i<<8), &val);
		owner()->adcTemperature.set(i, val);
	}
	if (ret)
		cerr << "Error CAEN_DGTZ_ReadRegister" << endl;

	close();
}

/*!
 * \brief Enable Coincidence command implementation
 */
void RealCAENCfdDriver::useCoincidence() {

	open();

	uint32_t val;
	uint32_t addr = 0x1080;

	ret |= CAEN_DGTZ_ReadRegister(m_boardHandle, addr, &val);
	if (ret)
		cerr << "Error CAEN_DGTZ_ReadRegister" << endl;

	if (owner()->enableCoincidence() == 1) {
		val = val | (1 << 19);
	}

	ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, addr, val);
	if (ret)
		cerr << "Error CAEN_DGTZ_WriteRegister" << endl;

	addr = 0x8080;
	if (owner()->enableCoincidence() == 1) {
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, addr, 0);// signals from mother-board (inter-couple) are not used
		if (ret)
			cerr << "Error CAEN_DGTZ_WriteRegister" << endl;
	}

	addr = 0x10A0;
	if (owner()->enableCoincidence() == 1) {
		ret |= CAEN_DGTZ_ReadRegister(m_boardHandle, addr, &val);
		if (ret)
			cerr << "Error CAEN_DGTZ_ReadRegister" << endl;

		val = val | 0x00000077;						// enable couple trgout // enable trg validation input
		// trigger propagation between the channels in the couple // couple trgout = OR (NOTE: doesn't reflect the acq. trigger!!!)

		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, addr, val);
		if (ret)
			cerr << "Error CAEN_DGTZ_WriteRegister" << endl;

	}

	// Channel Trigout Width (it determines the coincidence window).
	if (owner()->enableCoincidence() == 1) {
		val = owner()->coincidenceWindow();
		addr = 0x1084;
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, addr, val);
		if (ret)
			cerr << "Error CAEN_DGTZ_WriteRegister" << endl;

		uint32_t addr = 0x8184;
		ret |= CAEN_DGTZ_ReadRegister(m_boardHandle, addr, &val);
		if (ret)
			cerr << "Error CAEN_DGTZ_ReadRegister" << endl;

		val = (val & 0x3FF) + owner()->coincidenceMask();
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, addr, val);
		if (ret)
			cerr << "Error CAEN_DGTZ_WriteRegister" << endl;
	}

	close();
}

/*!
 * \brief Set CFD mode parameters
 */
void RealCAENCfdDriver::setCfdParam() {

	for (int32 i=0; i<owner()->nbChannels(); ++i) {

		if (owner()->cfdActive.get(i) == 1) {
			ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x10D0 + (i<<8), owner()->cfdSignalZero.get(i));	// input baseline level in ADC counts
			// Activate CFD mode
			registerSetBits(0x1080 + (i<<8), 6, 6, 0);
			registerSetBits(0x1080 + (i<<8), 17, 17, 1);
			// CFD fraction
			registerSetBits(0x10A0 + (i<<8), 19, 19, 0);
			registerSetBits(0x10A0 + (i<<8), 12, 13, owner()->cfdFraction.get(i));
			ret = CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1058 + (i<<8), owner()->cfdDelay.get(i));
			owner()->cfdActiveStr.set(i, "Active");

		} else {
			registerSetBits(0x1080 + (i<<8), 6, 6, 0);
			registerSetBits(0x1080 + (i<<8), 17, 17, 0);
			registerSetBits(0x10A0 + (i<<8), 12, 13, 0);
			registerSetBits(0x10A0 + (i<<8), 19, 19, 0);
			owner()->cfdActiveStr.set(i, "Inactive");
		}

		// Set Extra Word = CFD zero crossing samples
		registerSetBits(0x10A0 + (i<<8), 8, 10, 2);   // Extra Word = extended Tstamp + fine Tstamp
	}
}

void RealCAENCfdDriver::registerSetBits(uint16_t addr, int32 startBit, int32 endBit, int32 val) {

	uint32_t mask = 0;
	uint32_t reg = 0;

	// broadcast access to channel individual registers (loop over channels)
	if (((addr & 0xFF00) == 0x8000) && (addr != 0x8000) && (addr != 0x8004) && (addr != 0x8008)) {
		for(int32 ch=0; ch<owner()->nbChannels(); ++ch) {
			ret |= CAEN_DGTZ_ReadRegister(m_boardHandle, 0x1000 | (addr & 0xFF) | (ch << 8), &reg);
			for(int32 i=startBit; i<=endBit; i++)
				mask |= 1 << i;
			reg = (reg & ~mask) | ((val<<startBit) & mask);
			ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1000 | (addr & 0xFF) | (ch << 8), reg);
		}
	} else { // access to channel individual register or mother board register
		ret |= CAEN_DGTZ_ReadRegister(m_boardHandle, addr, &reg);
		for(int32 i=startBit; i<=endBit; i++)
			mask |= 1<<i;
		reg = (reg & ~mask) | ((val<<startBit) & mask);
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, addr, reg);
	}
	if (ret)
		cerr << "Error registerSetBit" << endl;
}

int32 RealCAENCfdDriver::LockTempCalibration_x730(uint32_t ch) {

	uint32_t lock, ctrl;

	// enter engineering functions
	ret |= WriteSPIRegister(ch, 0x7A, 0x59);
	ret |= WriteSPIRegister(ch, 0x7A, 0x1A);
	ret |= WriteSPIRegister(ch, 0x7A, 0x11);
	ret |= WriteSPIRegister(ch, 0x7A, 0xAC);

	// read lock value
	ret |= ReadSPIRegister (ch, 0xA7, &lock);
	// write lock value
	ret |= WriteSPIRegister(ch, 0xA5, lock);

	// enable lock
	ret |= ReadSPIRegister (ch, 0xA4, &ctrl);
	ctrl |= 0x4;  // set bit 2
	ret |= WriteSPIRegister(ch, 0xA4, ctrl);

	ret |= ReadSPIRegister (ch, 0xA4, &ctrl);

	return ret;
}

int32 RealCAENCfdDriver::ReadSPIRegister(uint32_t ch, uint32_t address, uint32_t *value) {

	uint32_t SPIBusy = 1;
    uint32_t SPIBusyAddr        = 0x1088 + (ch<<8);
    uint32_t addressingRegAddr  = 0x80B4;
	uint32_t valueRegAddr       = 0x10B8 + (ch<<8);
	uint32_t val                = 0x0000;

    while(SPIBusy) {
        if((ret = CAEN_DGTZ_ReadRegister(m_boardHandle, SPIBusyAddr, &SPIBusy)) != CAEN_DGTZ_Success)
            return CAEN_DGTZ_CommError;
        SPIBusy = (SPIBusy>>2)&0x1;
	    if (!SPIBusy) {
            if((ret = CAEN_DGTZ_WriteRegister(m_boardHandle, addressingRegAddr, address)) != CAEN_DGTZ_Success)
                return CAEN_DGTZ_CommError;
            if((ret = CAEN_DGTZ_ReadRegister(m_boardHandle, valueRegAddr, value)) != CAEN_DGTZ_Success)
                return CAEN_DGTZ_CommError;
	    }
        usleep(1000);
    }
    return CAEN_DGTZ_Success;
}

int32 RealCAENCfdDriver::WriteSPIRegister(uint32_t ch, uint32_t address, uint32_t value) {

	uint32_t SPIBusy = 1;

    uint32_t SPIBusyAddr        = 0x1088 + (ch<<8);
    uint32_t addressingRegAddr  = 0x80B4;
    uint32_t valueRegAddr       = 0x10B8 + (ch<<8);

    while(SPIBusy) {
        if((ret = CAEN_DGTZ_ReadRegister(m_boardHandle, SPIBusyAddr, &SPIBusy)) != CAEN_DGTZ_Success)
            return CAEN_DGTZ_CommError;
        SPIBusy = (SPIBusy>>2)&0x1;
	    if (!SPIBusy) {
            if((ret = CAEN_DGTZ_WriteRegister(m_boardHandle, addressingRegAddr, address)) != CAEN_DGTZ_Success)
                return CAEN_DGTZ_CommError;
            if((ret = CAEN_DGTZ_WriteRegister(m_boardHandle, valueRegAddr, value)) != CAEN_DGTZ_Success)
                return CAEN_DGTZ_CommError;
	    }
        usleep(1000);
    }
    return CAEN_DGTZ_Success;
}

}
