/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "RealQE65000Driver.h"
#include <fstream>
#define LINUX
#include <iomanip>
#include "api/seabreezeapi/SeaBreezeAPI.h"
#include "api/SeaBreezeWrapper.h"

using namespace std;

namespace oceanoptics {

RealQE65000Driver::RealQE65000Driver(QE65000Driver* owner) :
		QE65000State(owner) {

	spectrometerId = 0;
	deviceId = 0;
}

RealQE65000Driver::~RealQE65000Driver() {
    sbapi_shutdown();

}

void RealQE65000Driver::init() {

	owner()->commandProgression = 10;
	//Get deviceID necessary to use library  methods

	sbapi_initialize();
	sbapi_probe_devices();

	//Retrieve necessary ids
	getDeviceID();
	getSpectroID();
	int32 error = 0;
	int32 result = sbapi_open_device(deviceId, &error);
	if (result == 0) {
		int32 error = 0;
		owner()->size = sbapi_spectrometer_get_formatted_spectrum_length(deviceId, spectrometerId, &error);
		owner()->openError = false;

	} else {
		owner()->openError = true;
	}
	sbapi_close_device(deviceId, &error);

	owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
}

void RealQE65000Driver::start() {
	int32 error = 0;
	owner()->openError = false;
	//Delete current arrays
	if (owner()->xData()) {
		delete[] owner()->xData();
	} else if (owner()->yData())  {
		delete[] owner()->yData();
	}

	//Modification of 17/08/2016 -> really slow access to the following methods. Call them only in init method.
//	//Get deviceID necessary to use library  methods
//	sbapi_initialize();
//	sbapi_probe_devices();
//
//	//Retrieve necessary ids
//	getDeviceID();
//	getSpectroID();
	//End of modification of 17/08/2016

	//Open device
	int32 result = sbapi_open_device(deviceId, &error);
	if (result == 0) {
		//Get size
		int32 error = 0;
		owner()->size = sbapi_spectrometer_get_formatted_spectrum_length(deviceId, spectrometerId, &error); //needed here??
		//Set integration time
		setIntegrationTime(owner()->integrationTime());

		//Retrieve data
		getFormattedSpectrum();
		getWavelength();

		owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;

		sbapi_close_device(deviceId, &error);

	} else {
		cout << "error opening the device" << endl;
		owner()->openError = true;
		//throw CannotOpenDevice(owner()->getType(), owner()->getName(), owner(), "Cannot open QE65000 spectrometer");
	}

}

void RealQE65000Driver::stop() {
	owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
}

/**
 * Check the status of the spectrometer: plugged or unplugged. NOT USED !
 */
void RealQE65000Driver::readStatus() {
	//time consuming methods
//	sbapi_initialize();
//	sbapi_probe_devices();
	//end of time consuming methods

	int32 numberOfDevices = sbapi_get_number_of_device_ids();
	if (numberOfDevices != 0) {
		owner()->status.update(QE65000Driver::PLUGGED);

	} else {
		owner()->status.update(QE65000Driver::UNPLUGGED);
	}
}

void RealQE65000Driver::getDeviceID() {
	//There should be only ONE device
	int32 numberOfDevices = sbapi_get_number_of_device_ids(); //Provides the number of devices that have either been probed or manually specified
	if (numberOfDevices == 0) {
		// error uhuuu
		owner()->status.update(QE65000Driver::UNPLUGGED);
		owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
		return;

	} else if (numberOfDevices > 1) {
		// we will only use one by now...
		cout << "more than one device" << endl;
		owner()->status.update(QE65000Driver::PLUGGED);
	}

	sbapi_get_device_ids(&deviceId, 1); //Only one device...
	cout << "getting device id " << deviceId << endl;
}

void RealQE65000Driver::getSpectroID() {
	//get spectrometer id
	int32 error = 0;
	//need to get spectrometer feature IDs --> it is the same that opening it...
	int32 spectroNumbers = sbapi_get_number_of_spectrometer_features(deviceId, &error);

	if (spectroNumbers == 0) {
		// log error
		owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
		return;
	} else if (spectroNumbers > 1) {
		// log warning : only 1 used
	}

	sbapi_get_spectrometer_features(deviceId, &error, &spectrometerId, 1); //Only one device...
}

void RealQE65000Driver::setIntegrationTime(int32 integrationTime) {
	int32 error = 0;

	//NOTE: integration time must be bigger than minimum integration time
	int32 minimumIntegrationTime = sbapi_spectrometer_get_minimum_integration_time_micros(deviceId, spectrometerId, &error);

	if (minimumIntegrationTime > integrationTime) {
		integrationTime = minimumIntegrationTime;
	}
	cout << "in real qe65000 driver setting integration time to " << integrationTime << " minIntegration time " << minimumIntegrationTime <<endl;
	//api: spectrometerSetIntegrationTimeMicros (long deviceID, long featureID, int *errorCode, unsigned long integrationTimeMicros)
	sbapi_spectrometer_set_integration_time_micros(deviceId, spectrometerId, &error, integrationTime);
}

/**
 * Acquire a spectrum. In this mode, auto-nulling should be automatically performed for devices that support it.
 */
void RealQE65000Driver::getFormattedSpectrum() {
	int32 error = 0;

	float64* doubleBuffer = (float64 *) calloc(owner()->size(), sizeof(float64));
	sbapi_spectrometer_get_formatted_spectrum(deviceId, spectrometerId, &error, doubleBuffer, owner()->size());

	//Spectrum = Y data
	owner()->yData.set(doubleBuffer);
	owner()->yData.setSize((int64) owner()->size());
}

/*
 * Get Wavelengths or x data
 */
void RealQE65000Driver::getWavelength() {
	int32 error = 0;

	int32 length = sbapi_spectrometer_get_formatted_spectrum_length(deviceId, spectrometerId, &error);
	float64* valuesBuffer = (float64 *) calloc(owner()->size(), sizeof(float64));
	sbapi_spectrometer_get_wavelengths(deviceId, spectrometerId, &error, valuesBuffer, owner()->size());

	owner()->xData.set(valuesBuffer);
	owner()->xData.setSize((int64) owner()->size());
}

int32 RealQE65000Driver::getSize() {
	//Get deviceID necessary to use library  methods
	sbapi_initialize();
	sbapi_probe_devices();

	//Retrieve necessary ids
	getDeviceID();
	getSpectroID();
	int32 error = 0;
	int32 result = sbapi_open_device(deviceId, &error);
	if (result == 0) {
		int32 error = 0;
		owner()->size = sbapi_spectrometer_get_formatted_spectrum_length(deviceId, spectrometerId, &error);

	}
	sbapi_close_device(deviceId, &error);
	return owner()->size();
}

/**
 * Manipulate the shutter. It seems to me that this feature is not available
 */
void RealQE65000Driver::setShutter() {
	int32 error = 0;
	char opened = 0; //If 0 -> shutter closed, if 1 -> shutter opened
	int32 numberOfShutters = sbapi_get_number_of_shutter_features(deviceId, &error);

	if (numberOfShutters != 0) {
		int64* shutterIds = (int64 *) calloc(numberOfShutters, sizeof(int64));
		int32 numberOfShuttersWithFeature = sbapi_get_shutter_features(deviceId, &error, shutterIds, numberOfShutters);

		//Iterate over the shutter features
		for (int32 i = 0; i < numberOfShuttersWithFeature; i++) {
			//open shutter
			sbapi_shutter_set_shutter_open(deviceId, shutterIds[i], &error, 1);
			cout << "shutter open!" << endl;

			//close shutter
			sbapi_shutter_set_shutter_open(deviceId, shutterIds[i], &error, 0);
			cout << "shutter close!" << endl;
		}
	}
}

}
